@@ -13,7 +13,6 @@
 
 import contextlib
 import functools
-import inspect
 import operator
 import threading
 import warnings
@@ -27,6 +26,7 @@ from oslo_db import exception
 from oslo_db import options
 from oslo_db.sqlalchemy import engines
 from oslo_db.sqlalchemy import orm
+from oslo_db.sqlalchemy import utils
 
 
 class _symbol(object):
@@ -149,7 +149,8 @@ class _TransactionFactory(object):
             'thread_checkin': _Default(True),
             'json_serializer': _Default(None),
             'json_deserializer': _Default(None),
-            'logging_name': _Default(None)
+            'logging_name': _Default(None),
+            'connection_parameters': _Default(None)
         }
         self._maker_cfg = {
             'expire_on_commit': _Default(False),
@@ -164,15 +165,15 @@ class _TransactionFactory(object):
             'on_engine_create': [],
         }
 
-        # other options that are defined in oslo.db.options.database_opts
-        # but do not apply to the standard enginefacade arguments
-        # (most seem to apply to api.DBAPI).
+        # other options that are defined in oslo_db.options.database_opts
+        # or oslo_db.concurrency.tpool_opts but do not apply to the standard
+        # enginefacade arguments (most seem to apply to api.DBAPI).
         self._ignored_cfg = dict(
             (k, _Default(None)) for k in [
                 'db_max_retries', 'db_inc_retry_interval',
                 'use_db_reconnect',
                 'db_retry_interval', 'min_pool_size',
-                'db_max_retry_interval', 'backend'])
+                'db_max_retry_interval', 'backend', 'use_tpool'])
 
         self._started = False
         self._legacy_facade = None
@@ -219,6 +220,9 @@ class _TransactionFactory(object):
         :param connection_debug: engine logging level, defaults to 0. set to
          50 for INFO, 100 for DEBUG.
 
+        :param connection_parameters: additional parameters to append onto the
+         database URL query string, pass as "param1=value1&param2=value2&..."
+
         :param max_pool_size: max size of connection pool, uses CONF for
          default
 
@@ -393,7 +397,8 @@ class _TransactionFactory(object):
             self._start()
         if mode is _WRITER:
             return self._writer_engine.connect()
-        elif self.synchronous_reader or mode is _ASYNC_READER:
+        elif mode is _ASYNC_READER or \
+                (mode is _READER and not self.synchronous_reader):
             return self._reader_engine.connect()
         else:
             return self._writer_engine.connect()
@@ -408,7 +413,8 @@ class _TransactionFactory(object):
             kw['bind'] = bind
         if mode is _WRITER:
             return self._writer_maker(**kw)
-        elif self.synchronous_reader or mode is _ASYNC_READER:
+        elif mode is _ASYNC_READER or \
+                (mode is _READER and not self.synchronous_reader):
             return self._reader_maker(**kw)
         else:
             return self._writer_maker(**kw)
@@ -843,7 +849,8 @@ class _TransactionContextManager(object):
         new = self._clone()
         new._root = new
         new._root_factory = self._root_factory._create_factory_copy()
-        assert not new._factory._started
+        if new._factory._started:
+            raise AssertionError('TransactionFactory is already started')
         return new
 
     def patch_factory(self, factory_or_manager):
@@ -869,7 +876,8 @@ class _TransactionContextManager(object):
             raise ValueError(
                 "_TransactionContextManager or "
                 "_TransactionFactory expected.")
-        assert self._root is self
+        if self._root is not self:
+            raise AssertionError('patch_factory only works for root factory.')
         existing_factory = self._root_factory
         self._root_factory = factory
 
@@ -970,7 +978,7 @@ class _TransactionContextManager(object):
 
     def __call__(self, fn):
         """Decorate a function."""
-        argspec = inspect.getargspec(fn)
+        argspec = utils.getargspec(fn)
         if argspec.args[0] == 'self' or argspec.args[0] == 'cls':
             context_index = 1
         else:
